function [decodedAudio,sampleRate] = DRMread(strFileName,startSec,stopSec)
%------------------------------------------------------------------------+
%  PROJECT :   DRM - Decoding Morphy Richards *.DRM storage files
%------------------------------------------------------------------------%
%
%   SYNTAX:    [SAMPLES, SAMPLERATE] = ...
%                   DRMread(FILENAME [,START [,STOP]])
%   
%   MODULE  :   DRMread.m   [Function]
%
%   DEPEND. :   drm_aacdecode.dll   .. mex-wrapper for 'libfaad2.dll'
%               libfaad2.dll        .. FAAD2 decoder library
%               bytes2bits.dll      .. conversion function in 'C'
% 
%   DATE    :   07. Sept. 2007
% 
%   AUTHOR  :   Ferenc Kraemer, TU-Ilmenau, Fraunhofer IDMT
% 
%   ABSTRACT:   Function to decode *.DRM files of the 'Morphy Richards DRM
%               Radio' utilizing the 'Free Advanced Audio Decoder v.2'
%               (FAAD2) (http://sourceforge.net/projects/faac/).
%               The routines for extracting the AAC-Frames from the DRM
%               Audio-Super-Frames (ASF) where taken from the DIORAMA 
%               software in Ref.[1].
%               With minor changes the inherent mehtods can also be used to
%               decode the ASF-Frames from MDI/DCP-Paket streams.
%
%   REFERENCE:  [1] DIORAMA DRM receiver software for MATLAB;
%                   http://nt.eit.uni-kl.de/forschung/diorama/
%
%   PARAMETER:  
%       OUTPUT: decodedAudio   .. audio samples of the decoded *.DRM file
%                                 [n,1]-vector (mono) 
%                                 [n,2]-matrix (stereo)
%               sampleRate     .. sampling frequency of decoded audio in
%                                 Hertz [Hz]
%
%       INPUT : strFileName    .. string containing the input name of the 
%                                 file to be decoded. E.g: 'REC00178.DRM'
%               startSec       .. optional start parameter of decoding
%                                 onset in seconds [s]
%               stopSec        .. optional stop parameter for decoding [s]
%                    
%   HISTORY :   (fk-07.09.2007) - Creation of this function
%
%   PROBLEMS:   (fk-07.09.2007) - The 'drm_aacdecode.dll' ('lobfaad2.dll')
%                   seems to need some kind of init and stop mechanism not 
%                   provided here. This leads to a lock of those libraries
%                   to the current working directory. I.e. a function call
%                   from a different location may not work! Restarting
%                   MATLAB solves the Problem ;).
%               (fk-07.09.2007) - The Interface Header leading each
%                   ASF-Frame which contains parameters needed for decoding
%                   is read only read once at the beginning. So rescaling of 
%                   the DRM content steam within the *.DRM file will cause
%                   a crash error.
%               (fk-07.09.2007) - So far only Equal Error Protection (EEP)
%                   can be decoded!
%               (fk-10.09.2007) - SBR seems not to be working
%
%------------------------------------------------------------------------/
if ~ispc
    disp(['Sorry, but this function only runs under WIN32 due to'...
    ' library dependencies (''bytes2bits.dll'' & ''drm_aacdecode.dll'')']);
    return;
end
%% <Parameter> ----------------------------------------------------------+   
        
    % default start/stop parameter
        % extracting start time in seconds
        if nargin < 2
            startSec = 0;
        end
        % extracting stop time in seconds
        if nargin < 3
            stopSec  = inf;
        end
        
    % number of bytes in the Morphy Richards *.DRM file header
        nByteHeader_Morphy = 40; % [byte]

    % number of bytes for the ASF Interface Header
        uLongB = 4; % bytes for unsigned long
        nByteHeader_Interface = 11*uLongB; % 44byte = 11*unsinged long

        
% </Parameter> ----------------------------------------------------------/
%% <BODY> ---------------------------------------------------------------+
        
    %% initialization ................................................ %
        % Audio-Super-Frame indices corresponding to start/stop times
        startASF = floor(startSec/.4)+1;    % 400 ms/ASF-Frame
        stopASF  = ceil(stopSec/.4)+1; 
    
    %% main procedure ................................................ %
        % open file for binary reading
        fid = fopen(strFileName, 'rb');
        
        % . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . %
        % extract Morphy Richards header
        MorphyHeader = fread(fid, nByteHeader_Morphy, '*char')';
        disp('* - Morphy Richards header------------------------- *');
        disp(MorphyHeader);
            
        % . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . %    
        % extract interface-header
        IntHeader = fread(fid, nByteHeader_Interface,'uint8');
        IntHeader = reshape(IntHeader,uLongB,nByteHeader_Interface/uLongB);
        hex2uLongB = 16.^[6:-2:0]';
        hex2uLongP = 16.^[7:-2:1]';
        IntHead = zeros(size(IntHeader,2),1);
        for id=1:size(IntHeader,2)
            base = rem(IntHeader(:,id),16).*hex2uLongB;
            pw = floor(IntHeader(:,id)./16).*hex2uLongP;
            IntHead(id) = sum(base+pw); 
        end
        InterfaceHeader.lengthInterfaceFrame    = IntHead(1);
        InterfaceHeader.versionID               = IntHead(2);
        InterfaceHeader.coderSamplingRate       = IntHead(3);
        InterfaceHeader.bitrate                 = IntHead(4);
        InterfaceHeader.coderID                 = IntHead(5);
        InterfaceHeader.audioMode               = IntHead(6);
        InterfaceHeader.SBROnOffFlag            = IntHead(7);
        InterfaceHeader.lengthOfAudioFrame      = IntHead(8);
        InterfaceHeader.lengthHigherProtected   = IntHead(9);
        InterfaceHeader.UEPUsedFlag             = IntHead(10);
        InterfaceHeader.errorIndicator          = IntHead(11);
        disp('* - ASF Interface header--------------------------- *');
        InterfaceHeader
        
        if InterfaceHeader.UEPUsedFlag || ...
                InterfaceHeader.lengthHigherProtected > 0
            error('UEP-Mode not supported!');
        end
        
        % . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . %    
        % extract AAC-Frames from ASF-Frames
        
        % skip leading Morphy Richards header
        fseek(fid, nByteHeader_Morphy, 'bof');
        
        % read whole content from file
        InterfaceData = fread(fid, 'uint8');
        fclose(fid);
    
        % length of interface frames [byte]
        lFrame = InterfaceHeader.lengthInterfaceFrame;
        % length of ASFs [byte]
        lFrameASF = InterfaceHeader.lengthOfAudioFrame;
        sampleRate = InterfaceHeader.coderSamplingRate;
        if sampleRate == 24000
            num_frames = 10; % AAC-Frames per ASF
        else % 12kHz !
            num_frames = 5;
        end
        % number of bytes in the ASF header (6@12kHz or 14@24kHz)
        numByteASFheader = ceil((num_frames-1)*12/8);
        % number of audio_super_frames alltogether
        nFrames = floor(length(InterfaceData)./lFrame);
        % padding with zeros if necessary:
        if length(InterfaceData) > (lFrame*nFrames)
            nFrames = nFrames + 1;
            DataN = zeros(lFrame.*nFrames,1);
            DataN(1:length(InterfaceData)) = InterfaceData;
            InterfaceData = DataN;        
        end
    
        %%% extracting aac_audio_frames from audio_super_frames, Ref.[1]
        % final number of frames to extract
%         nFramesExt = min(nFrames,stopASF)-startASF;
        % prealloc vars
%         Audio_Super_Frames  = zeros(nFramesExt,lFrameASF);
%         FrameLengths        = zeros(nFramesExt,num_frames);
        AudioFrames = {};
        id = 1;
        for frame=startASF:min(nFrames,stopASF)
            % extract ONLY the ASFs. Interface header will be skiped here
            audio_super_frame = InterfaceData( (frame-1).*lFrame + ...
                nByteHeader_Interface+(1:lFrameASF));
            
            % this is a bit strange and should be altered in some way
            if lFrameASF == 870
                audio_super_frame(end) = [];
            end
            
            % audio_super_frame_header:
            asf_header = audio_super_frame(1:numByteASFheader);
            asfh_bits  = bytes2bits(uint8(asf_header));        
            % extracting frame lentghs of aac_audio_frames from header
            frame_borders = reshape(asfh_bits(1:12*(num_frames-1)),...
                12,num_frames-1)'*(2.^[11:-1:0]');
            

            % extracting aac_audio_frames from audio_super_frames
            aacBytes=audio_super_frame(numByteASFheader+num_frames+1:end);
            frame_length = [frame_borders;length(aacBytes)]...
                            - [0;frame_borders];
            % crc bytes needed for decoding
            crc = audio_super_frame(numByteASFheader+(1:num_frames));
            previousborder = 0;
            for iid=1:num_frames
                if ((previousborder >= 0) && (previousborder + ...
                        frame_length(iid) <= length(aacBytes) ) && ...
                        (frame_length(iid) > 0))
                    AACframes{iid} = [crc(iid); ...
                     aacBytes(previousborder + [1:frame_length(iid)])];
                else
                    AACframes{iid} = 0;
                end
                previousborder = previousborder + frame_length(iid);
            end
            
            % save AAC-Frames per ASF for later decoding
            AudioFrames = [AudioFrames,AACframes];
            
%             FrameLengths(id,:) = 1 + frame_length; % +1 crc!
%             Audio_Super_Frames(id,:) = audio_super_frame;
            
            id = id+1;
        end
        numAACFrames = length(AudioFrames);
        
        % . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . %  
        % write audio_super_frames back to file
%         DataIn = reshape(Audio_Super_Frames',size(Audio_Super_Frames,1)...
%             .*size(Audio_Super_Frames,2),1);
%         strFileNameW = [strFileName '_mod.DRM'];
%         fid = fopen(strFileNameW, 'wb');
%         fwrite(fid, DataIn, 'uint8');
%         fclose(fid);

        % . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . %  
%         % write AudioFrames to file
%         fid = fopen([strFileName(1:end-4) '.aac'], 'wb');
%         for k=1:numAACFrames
%             fwrite(fid,AudioFrames{k},'uint8');
%         end
%         fclose(fid);
%         disp(['AAC data writen to: ' [strFileName(1:end-4) '.aac']]);
        % . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . %
        % docoding of AAC-Frames with FAAD2 (libfaad2.dll), Ref.[1]

        % audio mode: 0=AAC mono, 1=AAC parametric stereo, 2=AAC stereo
        % FAAD2: /* DRM channel definitions */
        % #define DRMCH_MONO          1
        % #define DRMCH_STEREO        2
        % #define DRMCH_SBR_MONO      3
        % #define DRMCH_SBR_STEREO    4
        % #define DRMCH_SBR_PS_STEREO 5
        
        audioMode  = InterfaceHeader.audioMode;
        sbrFlag = InterfaceHeader.SBROnOffFlag;
        switch audioMode
            case 0
                if sbrFlag
                    audioMode = 3;
                else
                    audioMode = 1;
                end
            case 1
                    audioMode = 5;  
            case 2
                if sbrFlag
                    audioMode = 4;                                 
                else
                    audioMode = 2;
                end
            otherwise
                error('Wrong audio mode.');
        end
        % number of samples per audio frame (960 or 1920)
        if sbrFlag
            numSamples = sampleRate*.4/num_frames*2;        
        else
            numSamples = sampleRate*.4/num_frames;
        end               
        % prealloc var
        if audioMode == 1 || audioMode == 3
            decodedAudio = zeros(1,numSamples*numAACFrames);
        else
            decodedAudio = zeros(2,numSamples*numAACFrames);        
        end
        % decode AAC-Frames
        for k=1:numAACFrames,     
            if AudioFrames{k} == 0
                if audioMode == 1 || audioMode == 3 % mono
                    blockdec = zeros(1,numSamples);
                else
                    blockdec = zeros(2,numSamples);
                end
            elseif audioMode == 1 || audioMode == 3 % mono
                % FAAD function call
                blockdec = ...
                    sum(double(drm_aacdecode(uint8(AudioFrames{k}),...
                        sampleRate,audioMode)),1)/2;
%                 blockdec = ...
%                         (double(drm_aacdecode(uint8(AudioFrames{k}),...
%                         sampleRate,audioMode)));
            else
                blockdec = double(drm_aacdecode(uint8(AudioFrames{k})',...
                    sampleRate,audioMode));
            end
            % in case of error fill with zeros
            if isempty(blockdec)
                if audioMode == 1 || audioMode == 3 % mono
                    blockdec = zeros(1,numSamples);
                else
                    blockdec = zeros(2,numSamples);
                end
            end
            decodedAudio(:,(k-1)*numSamples+(1:numSamples)) = blockdec;
        end
        drm_aacdecode([],sampleRate,audioMode);	%close decoder
    
        % scale down from 16Bit integer
        decodedAudio = decodedAudio.';
        decodedAudio = decodedAudio./2^15;
        
        % check clipping
%         if max(max(abs(decodedAudio))) > 1
%             decodedAudio = decodedAudio./max(max(abs(decodedAudio)));
%             disp('INFO_DRMread: Normalized due to cliping!');
%         end
        
        % set output sampling rate
        if sbrFlag
            sampleRate = sampleRate*2;
        end

        disp(['...' num2str(size(decodedAudio,1)/sampleRate) ...
          ' seconds of audio processed.' ]);
    
%     sound(decodedAudio,sampleRate)
    
    % . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . %
    % write decoded audio file to disc    
    wavwrite(decodedAudio,sampleRate,[strFileName(1:end-4) '.wav']);
    disp(['Audio written to: ' [strFileName(1:end-4) '.wav']]);
    
% </BODY> ---------------------------------------------------------------/